#include <assert.h>
#include <time.h>
#include "base.h"


static verylong ts_n = 0;

static void CheckTs(ostream& checkfile)
{
	checkfile << ts_n;
}

/*
 * This function calculates the aliquot sequence of n until it terminates.
 * There are sequences that never terminate: perfect and amicable numbers,
 * and possibly sequences that grow beyond all bounds.
 * Hence the loop never terminates for certain n.
 */
void ts(verylong n)
{
	IntList	l;

	zcopy(n, &ts_n);

	while (zscompare(ts_n, 276) >= 0)
	{
		// Output only looks good when trace_level = 0.
		cout << ts_n << " =";
		factorsum(ts_n, 1000000, &ts_n, l);
		cout << l << endl;
		if (checkpoint)
			WriteCheckMark(CheckTs);
		l.Clear();
	}

} // ts





static void print_factors(verylong n)
{
	IntList	l;

	ushort s = factor(n, 0, 1000, l);
	assert(s == NumPrime);
	assert(l.Verify(n));

	cout << ": " << n << " =" << l << endl;

} // print_factors





static const int	MaxCycle = 256;
// We are guaranteed to find all cycles where the quotient of last and
// first number doesn't exceed MaxQuotient.
static const ulong	MaxQuotient = 100;
static verylong		friend_n = 0;

static inline long MaxTrialDiv(verylong n)
{
	// This returns 5520 when n is 1.3e10, and 10800 when n is 7e11.
	long t = 880 * z2log(n) - 23520;
	return t < 5000 ? 5000 : t;
}

static void CheckFriendSearch(ostream& checkfile)
{
	checkfile << friend_n;
}

static void PrintTimestamp()
{
	time_t t = time(NULL);
	cout << ctime(&t);
}

static void SearchCycle(verylong s, verylong* r, int t)
{
	int i;

	for (i = MaxCycle; --i >= 0; )
		if (zcompare(r[i], s) == 0)
			break;

	if (i < 0)
	{
		cout << friend_n << ": Cycle too long" << endl << flush;
	}
	else
	{
		PrintTimestamp();
		cout << "  ...";
		if (t - i != 2)
		{
			for ( ; i < t; i++)
				cout << r[i] << " ";
		}
		else
		{
			if (zcompare(r[i], r[i + 1]) < 0)
			{
				cout << r[i];
				cout << " " << r[i + 1];
			}
			else
			{
				cout << r[i + 1];
				cout << " " << r[i];
			}
		}
		cout << endl << flush;
	}

} // SearchCycle

/*
 * Searches for amicable numbers in a never terminating loop
 * Input:	start > 0
 *		onlyone: if true, only the number start is analyzed
 *
 * The search begins with start and continues endlessly. All aliquot cycles
 * up to length 255 will be found if the quotient of largest and smallest
 * number doesn't exceed MaxQuotient (very unlikely). Every sequence will
 * only be detected once.
 */
void FriendSearch(verylong start, int onlyone)
{
	static verylong s = 0;
	static verylong	r[MaxCycle];
	static verylong	limit = 0;
	static bool	rInitialized = false;
	IntList		l;
	const ulong	loops = 100000;
	ulong		count = loops - zsmod(start, loops) + 1;

	if (rInitialized == false)
	{
		for (int i = 0; i < MaxCycle; i++)
			r[i] = 0;
		rInitialized = true;
	}

	zcopy(start, &friend_n);
	zsmul(friend_n, MaxQuotient, &limit);
	if (zscompare(friend_n, 10000) < 0)
		zsmul(limit, 1000, &limit);

	while (1)
	{
		int t = 0;

		if (--count == 0)
		{
			if (trace_level > 1)
				cout << "  Test " << friend_n << endl;
			count = loops;
			zsmul(friend_n, MaxQuotient, &limit);
		}

		factorsum(friend_n, MaxTrialDiv(friend_n), &s, l);
		l.Clear();
		long c = zcompare(s, friend_n);
		if (c > 0)
		{
			do
			{
				if (t >= MaxCycle)
				{
					// Test if the last number appears
					// twice.
					SearchCycle(s, r, t);
					break;
				}

				if (zcompare(s, limit) > 0)
				{
					//cout << friend_n << ": Number becomes too big" << endl << flush;
					break;
				}

				zcopy(s, &r[t++]);
				factorsum(s, MaxTrialDiv(s), &s, l);
				l.Clear();
			} while ((c = zcompare(s, friend_n)) > 0);
		}

		if (c == 0)
		{
			// We have found a cycle.
			int i;
			for (i = t; --i >= 0 && zcompare(r[i], friend_n) != 0; )
				;
			assert(i == -1);
			PrintTimestamp();
			if (i < 0)	// Always true?
			{
				cout << friend_n << " ";
				i = 0;
			}
			for ( ; i < t; i++)
				cout << r[i] << " ";
			if (t == 1)
				cout << ratio(friend_n, r[0]);
			cout << endl;
			print_factors(friend_n);
			for (i = 0; i < t; i++)
				print_factors(r[i]);
			cout << flush;
		}

		if (onlyone)
			break;

		if (checkpoint)
			WriteCheckMark(CheckFriendSearch);

		zsinc(friend_n, 1);
	}

} // FriendSearch




/*
 * Searches for augmented amicable numbers in a never terminating loop
 * Input:	start > 0
 *		onlyone: if true, only the number start is analyzed
 *
 * The search begins with start and continues endlessly. All augmented aliquot
 * cycles up to length 255 will be found if the quotient of largest and smallest
 * number doesn't exceed MaxQuotient (very unlikely). Every sequence will
 * only be detected once.
 */
void AugmentedSearch(verylong start, int onlyone)
{
	static verylong s = 0;
	static verylong	r[MaxCycle];
	static verylong	limit = 0;
	static bool	rInitialized = false;
	IntList		l;
	const ulong	loops = 100000;
	ulong		count = loops - zsmod(start, loops) + 1;

	if (rInitialized == false)
	{
		for (int i = 0; i < MaxCycle; i++)
			r[i] = 0;
		rInitialized = true;
	}

	zcopy(start, &friend_n);
	zsmul(friend_n, MaxQuotient, &limit);
	if (zscompare(friend_n, 10000) < 0)
		zsmul(limit, 1000, &limit);

	while (1)
	{
		int t = 0;

		if (--count == 0)
		{
			if (trace_level > 1)
				cout << "  Test " << friend_n << endl;
			count = loops;
			zsmul(friend_n, MaxQuotient, &limit);
		}

		factorsum(friend_n, MaxTrialDiv(friend_n), &s, l);
		zsinc(s, 1);
		l.Clear();
		long c = zcompare(s, friend_n);
		if (c > 0)
		{
			do
			{
				if (t >= MaxCycle)
				{
					// Test if the last number appears
					// twice.
					SearchCycle(s, r, t);
					break;
				}

				if (zcompare(s, limit) > 0)
				{
					//cout << friend_n << ": Number becomes too big" << endl << flush;
					break;
				}

				zcopy(s, &r[t++]);
				factorsum(s, MaxTrialDiv(s), &s, l);
				zsinc(s, 1);
				l.Clear();
			} while ((c = zcompare(s, friend_n)) > 0);
		}

		if (c == 0)
		{
			// We have found a cycle.
			int i;
			for (i = t; --i >= 0 && zcompare(r[i], friend_n) != 0; )
				;
			assert(i == -1);
			PrintTimestamp();
			if (i < 0)	// Always true?
			{
				cout << friend_n << " ";
				i = 0;
			}
			for ( ; i < t; i++)
				cout << r[i] << " ";
			if (t == 1)
				cout << ratio(friend_n, r[0]);
			cout << endl;
			print_factors(friend_n);
			for (i = 0; i < t; i++)
				print_factors(r[i]);
			cout << flush;
		}

		if (onlyone)
			break;

		if (checkpoint)
			WriteCheckMark(CheckFriendSearch);

		zsinc(friend_n, 1);
	}

} // AugmentedSearch




/*
 * Searches for quasi amicable numbers in a never terminating loop
 * Input:	start > 0
 *		onlyone: if true, only the number start is analyzed
 *
 * The search begins with start and continues endlessly. All quasi aliquot
 * cycles up to length 255 will be found if the quotient of largest and smallest
 * number doesn't exceed MaxQuotient (very unlikely). Every sequence will
 * only be detected once.
 */
void QuasiSearch(verylong start, int onlyone)
{
	static verylong s = 0;
	static verylong	r[MaxCycle];
	static verylong	limit = 0;
	static bool	rInitialized = false;
	IntList		l;
	const ulong	loops = 100000;
	ulong		count = loops - zsmod(start, loops) + 1;

	if (rInitialized == false)
	{
		for (int i = 0; i < MaxCycle; i++)
			r[i] = 0;
		rInitialized = true;
	}

	zcopy(start, &friend_n);
	zsmul(friend_n, MaxQuotient, &limit);
	if (zscompare(friend_n, 10000) < 0)
		zsmul(limit, 1000, &limit);

	while (1)
	{
		int t = 0;

		if (--count == 0)
		{
			if (trace_level > 1)
				cout << "  Test " << friend_n << endl;
			count = loops;
			zsmul(friend_n, MaxQuotient, &limit);
		}

		factorsum(friend_n, MaxTrialDiv(friend_n), &s, l);
		zsdec(s, 1);
		l.Clear();
		long c = zcompare(s, friend_n);
		if (c > 0)
		{
			do
			{
				if (t >= MaxCycle)
				{
					// Test if the last number appears
					// twice.
					SearchCycle(s, r, t);
					break;
				}

				if (zcompare(s, limit) > 0)
				{
					//cout << friend_n << ": Number becomes too big" << endl << flush;
					break;
				}

				zcopy(s, &r[t++]);
				factorsum(s, MaxTrialDiv(s), &s, l);
				zsdec(s, 1);
				l.Clear();
			} while ((c = zcompare(s, friend_n)) > 0);
		}

		if (c == 0)
		{
			// We have found a cycle.
			int i;
			for (i = t; --i >= 0 && zcompare(r[i], friend_n) != 0; )
				;
			assert(i == -1);
			PrintTimestamp();
			if (i < 0)	// Always true?
			{
				cout << friend_n << " ";
				i = 0;
			}
			for ( ; i < t; i++)
				cout << r[i] << " ";
			if (t == 1)
				cout << ratio(friend_n, r[0]);
			cout << endl;
			print_factors(friend_n);
			for (i = 0; i < t; i++)
				print_factors(r[i]);
			cout << flush;
		}

		if (onlyone)
			break;

		if (checkpoint)
			WriteCheckMark(CheckFriendSearch);

		zsinc(friend_n, 1);
	}

} // QuasiSearch





static verylong carm_base = 0;
static verylong carm_n = 0;

static void CheckCarmichael(ostream& checkfile)
{
	checkfile << carm_n;
	checkfile << " " << carm_base;
}

/*
 * Algorithmus:
 * Suche alle Primzahlen in einem Block von n Zahlen
 * Fuer alle nicht-Primzahlen in diesem Block fuehre strong_test und
 * Fermat-Test zur Basis base durch
 */
void Carmichael(verylong start, verylong base)
{
	static verylong	sieve = 0;
	static verylong	temp2 = 0;
	static verylong	temp3 = 0;
	ulong		block_len = 10000;

	zcopy(start, &carm_n);
	zcopy(base, &carm_base);

	if (zeven(carm_n))
		zsinc(carm_n, 1);

	// Limitation of the block length so that the precondition for
	// e_sieve is met.
	if (zscompare(carm_n, 102) < 0)
	{
		zssub(carm_n, 1, &temp2);
		zmul(carm_n, temp2, &temp3);
		zsadd(temp3, 2, &temp2);
		block_len = ztouint(temp2);
	}

	while (1)
	{
		ulong index = 0;
		ulong end = block_len / 2;

		zuintoz(block_len, &temp2);
		e_sieve(carm_n, temp2, carm_n, &sieve);

		if (trace_level > 1)
			cout << "Test " << carm_n << endl;

		while (index < end)
		{
			// A separate programm should test the even numbers.
			if (zbit(sieve, index))
			{
				if (strong_test(carm_n, base) == NumProbPrime)
				{
					cout << carm_n;
					cout << " is strong pseudoprime to base " << base << endl;
				}
				else
				{
					zssub(carm_n, 1, &temp2);
					zexpmod_m_ary(base, temp2, carm_n, &temp3, 1);
					if (zscompare(temp3, 1) == 0)
					{
						cout << carm_n;
						cout << " is pseudoprime to base " << base << endl;
					}
				}
			}
			zsinc(carm_n, 2);
			index++;
		}

		if (checkpoint)
			WriteCheckMark(CheckCarmichael);
	}

} // Carmichael

