#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include "base.h"


/******************************************************************************
 *
 * Constants and variables
 *
 *****************************************************************************/

int trace_level = 5;		// Amount of debug output (0-15)


// Simple sieve to omit numbers that are divisible by a small prime.
// Not as powerful as the primeDiff array, but not limited to
// MaxUpperLimit.
const int	incr_period = 2310;
short		incr_list[incr_period];
int		incr_len = 0;		// List length

ulong* primeDiff = NULL;





ostream& operator << (ostream& o, verylong n)
{
	static int	bufferSize = 0;
	static char*	buffer = NULL;
	static verylong	temp = 0;
	double		len;
	
	if (zsign(n) == 0)
		len = 1;
	else if (zsign(n) > 0)
		len = zslog(n, 10);
	else
	{
		zcopy(n, &temp);
		znegate(&temp);
		len = zslog(temp, 10);
	}

	len += 3;
	if (len > bufferSize)
	{
		bufferSize = (int) len + 1024;
		buffer = (char *) realloc(buffer, bufferSize);
		if (buffer == NULL)
		{
			cout << "Out of memory in operator <<" << endl;
			return o;
		}
	}

	zswrite(buffer, n);
	o << buffer;
	return o;

} // <<





double ratio(verylong x, verylong y)
{
	static verylong	r = 0;
	static verylong	tmp = 0;
	double result;

	zsmul(x, 1000000, &tmp);
	zdiv(tmp, y, &r, &tmp);
	result = zdoub(r) / 1000000;

	return result;

} // ratio





/*
 * Calculates p^n + p^(n-1) + ... + p + 1
 * Input:	p any
 *		n >= 0
 */
void powersum(verylong p, ulong n, verylong* r)
{
	zone(r);			// *r = 1 (r is a pointer)

	while (n-- > 0)
	{
		zmulin(p, r);
		zsadd(*r, 1, r);	// r = p * r + 1
	}

} // powersum





/*
 * Berechnet ein Array, das die Differenzen aufeinanderfolgender Zahlen
 * enthaelt, die nicht durch 2,3,5,7 oder 11 teilbar sind. Die Laenge des
 * Arrays ist incr_len. Das Array ist ab dem 6. Element (Index 5)
 * zyklisch verwendbar.
 */
void create_incr_list()
{
	ulong i, h, l, d;

	i = 13;		/* erste Zahl nach 2,3,5,7,11 */
	h = incr_period + 13;
	l = incr_period / 2;	/* Produkt aller Primzahlen < 13 */
	d = i;		/* Abstand zwischen den letzen beiden Primzahlen */
	incr_len = 0;
	incr_list[incr_len++] = 1;
	incr_list[incr_len++] = 2;
	incr_list[incr_len++] = 2;
	incr_list[incr_len++] = 4;
	incr_list[incr_len++] = 2;

	while ((i += 2) < h)
	{
		if (gcd(i, l) == 1)
		{
			incr_list[incr_len++] = (short) (i - d);
			d = i;
		}
	}

	incr_list[incr_len++] = (short) (i - d);

} // create_incr_list





/*
 * Lucas sequence mod n with parameters p and q
 * u(0) = 0; u(1) = 1; u(n+2) = pu(n+1) - qu(n)
 * v(0) = 2; v(1) = p; v(n+2) = pv(n+1) - qv(n)
 * Input:	m > 0	Index of the elements to calculate (u(m) and v(m))
 *		n > 0 odd	Modulo
 *		p, q	Parameters (cf. theorem8)
 * Output:	u, v	Calculated sequence elements
 * Note: With q = 1, p = x we have the Lucas (Chebyshev) polynomial Vn(x).
 */
void lucas(verylong m, verylong n, long p, long q,
	verylong* u, verylong* v)
{
	static verylong	temp = 0;
	static verylong	temp1 = 0;
	static verylong	temp2 = 0;
	static verylong	temp3 = 0;
	static verylong	temp4 = 0;
	static verylong	temp5 = 0;

	assert(zsign(m) > 0);
	assert(zsign(n) > 0);

	if (zscompare(m, 1) == 0)
	{
		zone(u);
		zintoz(p, v);
		zsadd(n, 1, &temp5);
		zrshift(temp5, 1, &temp5);	// temp5 = (n + 1) / 2
	}
	else
	{
		verylong	h = 0;

		z2div(m, &h);		// h = m / 2
		if (z2mod(m) == 0)
		{
			lucas(h, n, p, q, u, v);
			// Formula: u = (u * v) % n
			// Formula: v = (v * v - 2 * expt_mod(q, h, n)) % n
			zintoz(q, &temp);
			zexpmod_m_ary(temp, h, n, &temp4, 1);
			zmulmod(*u, *v, n, u);
			zsq(*v, &temp);
			z2mul(temp4, &temp1);
			zsub(temp, temp1, &temp1);
			zdiv(temp1, n, &temp, v);	// Don't use zmod
		}
		else
		{
			lucas(h, n, p, q, u, v);
			// Formula: v = (((p * p - 4 * q) * u * temp1 + p * (temp1 * temp1)) / 2
			// 	- p * expt_mod(q, h, n)) % n
			// Formula: u = ((p * u + temp1) * temp1 / 2 - expt_mod(q, h, n)) % n
			// Since we are calculating mod n, the division by 2
			// cannot be done directly; instead we must multiply
			// with the inverse of 2 mod n, which is temp5 = (n+1)/2.
			zintoz(q, &temp);
			zexpmod_m_ary(temp, h, n, &temp4, 1);
			zmulmod(*v, temp5, n, &temp3);
			zsmul(*u, p * p - 4 * q, &temp);
			zsmul(*v, p, &temp2);
			zadd(temp, temp2, &temp);
			zmul(temp, temp3, &temp2);
			zsmul(temp4, p, &temp1);
			zsub(temp2, temp1, &temp1);
			zsmul(*u, p, &temp);
			zadd(temp, *v, &temp2);
			zmul(temp2, temp3, &temp);
			zsub(temp, temp4, &temp2);
			zdiv(temp2, n, &temp, u);
			zdiv(temp1, n, &temp, v);
		}

		zfree(&h);
	}

	if (trace_level > 7)
	{
		cout << m << ": ";
		cout << *u << " ";
		cout << *v << endl;
	}

} // lucas





/*
 * Lucas (Chebychev) sequence mod n
 */
void lucas2(ulong m, verylong n,
	verylong p, verylong* u, verylong* v)
{
	static verylong	temp = 0;
	static verylong	temp2 = 0;
	static verylong	temp3 = 0;
	static verylong	temp5 = 0;

	if (m == 1)
	{
		zone(u);
		zcopy(p, v);
		zsadd(n, 1, &temp5);
		zrshift(temp5, 1, &temp5);	// temp5 = (n + 1) / 2
	}
	else
	{
		verylong	temp1 = 0;

		if (m % 2 == 0)
		{
			lucas2(m / 2, n, p, u, v);
			// Formula: u = (u * v) % n;
			zmulmod(*u, *v, n, u);
			// Formula: v = (v * v - 2) % n;
			zsq(*v, &temp);
			zssub(temp, 2, &temp);
			zdiv(temp, n, &temp, v);
		}
		else
		{
			lucas2(m / 2, n, p, u, &temp1);
			// Formula: v = (((p * p - 4) * u * v1
			//   + p * (v1 * v1)) / 2 - p) % n;
			// Formula: u = ((p * u + v1) * v1 / 2 - 1) % n;
			// Since we are calculating mod n, the division by 2
			// cannot be done directly; instead we must multiply
			// with the inverse of 2 mod n, which is temp5 = (n+1)/2.
			zsq(p, &temp2);
			zsdec(temp2, 4);
			zmul(*u, temp2, &temp);
			zmul(temp, temp1, &temp2);
			zsq(temp1, &temp);
			zmul(temp, p, &temp3);
			zadd(temp2, temp3, &temp);
			zmul(temp, temp5, &temp2);
			zsub(temp2, p, &temp);
			zdiv(temp, n, &temp2, v);
			zmul(p, *u, &temp);
			zadd(temp, temp1, &temp2);
			zmul(temp2, temp1, &temp);
			zmul(temp, temp5, &temp2);
			zone(&temp);
			zsub(temp2, temp, &temp);
			zdiv(temp, n, &temp2, u);
		}

		zfree(&temp1);
	}

	if (trace_level > 7)
	{
		cout << m << ": ";
		cout << *u << " ";
		cout << *v << endl;
	}

} // lucas2





/*
 * Applies the Fermat pseudo prime test to n.
 * Input:	n > 0
 *		base >= 2
 */
ushort fermat_test(verylong n, verylong base)
{
	static verylong temp = 0;
	static verylong temp1 = 0;

	assert(zsign(n) > 0);
	assert(zscompare(base, 2) >= 0);

	zssub(n, 1, &temp);
	zexpmod_m_ary(base, temp, n, &temp1, 1);
	if (zscompare(temp1, 1) == 0)
		return NumProbPrime;
	else
		return NumComposite;

} // fermat_test





/*
 * Applies a pseudo prime test to n which is slightly better than
 * Fermat's test.
 * Input:	n > 0
 *		base >= 2
 */
ushort strong_test(verylong n, verylong base)
{
	static verylong	temp = 0;
	static verylong	temp1 = 0;

	assert(zsign(n) > 0);
	assert(zscompare(base, 2) >= 0);

	z2div(n, &temp);	// temp = n / 2 [ == (n - 1) / 2 if n odd]

	if (zscompare(base, 2) == 0)
		z2expmod(temp, n, &temp1);
	else
		zexpmod_m_ary(base, temp, n, &temp1, 1);

	if (zscompare(temp1, 1) == 0)
		return NumProbPrime;

	zsadd(n, -1, &temp);
	if (zcompare(temp1, temp) == 0)
		return NumProbPrime;

	return NumComposite;

} // strong_test





/*
 * Calculates the status of n (prime, composite, or probable prime). The latter
 * can mean that n is, by chance, a base 2 pseudoprime, or a proper prime.
 * Input:	n odd
 *		u > 0	n is known to have no factor up to u
 * Return:	Status of n
 */
ushort num_status(verylong n, ulong u)
{
	static verylong	temp = 0;

	assert(u > 0);

	zuintoz(u, &temp);
	zsqin(&temp);			// temp = u * u

	if (zcompare(n, temp) > 0)
	{
		ztwo(&temp);
		return strong_test(n, temp);
	}

	return NumPrime;

} // num_status





void add_cofactor(verylong n, IntList& r, ulong upper)
{
	static verylong	temp1 = 0;
	static verylong	temp2 = 0;

	r.Multiply(&temp2);
	zdiv(n, temp2, &temp2, &temp1);
	assert(zsign(temp1) == 0);	// Must be divisible without remainder

	if (zscompare(temp2, 1) > 0)
	{	
		// If cofactor > 1, add it.
		r.Add(temp2, 1, num_status(temp2, upper - 1));
	}

} // add_cofactor





/*
 * Sucht Faktoren von n durch direkte Division. Es wird angenommen, dass
 * kein Faktor <= l existiert, und der hoechste Faktor, der probiert wird,
 * ist u. Die Liste primeDiff wird ausgenutzt.
 * Der Algorithmus sollte sehr schnell sein, da der Divisor ein ulong ist.
 * Das Resultat wird im Array r zurueckgegeben.
 * Das Ende des Arrays ist mit count = 0 markiert.
 * Input:	n > 0
 *		l >= 0	n is known to have no factor up to l
 *		u	Limit of trial divisions
 * Output:	r	Factors found
 * Return:	Status of the greatest factor
 */
ushort factor_brute(verylong m, ulong l, ulong u, IntList& r)
{
	assert(zsign(m) > 0);
	assert(u <= MaxUpperBound);

	int	c;
	int	t = 0;
	static verylong n = 0;
	static verylong temp = 0;
	static verylong s = 0;
	ulong	s1;

	zcopy(m, &n);

	if (u >= l)
	{
		ulong lower = l;
		l = 2;
		while (l < lower)
			l += primeDiff[t++];
	}

	zsqrt(n, &s, &temp);		// Upper bound of trial divisions
	if (zscompare(s, u) > 0)
		s1 = u;			// Stop trial divisions at u
	else
		s1 = ztouint(s);

	while (l <= s1)
	{
		if (zsmod(n, l) == 0)
		{
			c = 1;		// Multiplicity of a factor
			zsdiv(n, l, &n);
			while (zsmod(n, l) == 0)
			{
				zsdiv(n, l, &n);
				c++;
			}
			r.Add(l, c, NumPrime);
			zsqrt(n, &s, &temp);		// Upper bound of trial divisions
			if (zscompare(s, u) > 0)
				s1 = u;			// Stop trial divisions at u
			else
				s1 = ztouint(s);
		}

		l += primeDiff[t++];
	}

	add_cofactor(m, r, l);

	// If n was 1, r will be empty.
	return r.IsEmpty() ? NumPrime : r.GetState();

} // factor_brute





verylong* max3(verylong* n1, verylong* n2, verylong* n3)
{
	verylong* p = n1;

	if (zcompare(*n2, *p) > 0)
		p = n2;
	if (zcompare(*n3, *p) > 0)
		p = n3;

	return p;
}





/*
 * Sucht Faktoren von n, n-1 und n+1 durch direkte Division. Es wird
 * angenommen, dass kein Faktor <= l existiert, und der hoechste Faktor,
 * der probiert wird, ist u. Die Liste primeDiff wird ausgenutzt.
 * Der Algorithmus sollte sehr schnell sein, da der Divisor ein ulong ist.
 * Das Resultat wird in den Arrays r1, r2 und r3 zurueckgegeben.
 * Das Ende der Arrays ist mit count = 0 markiert.
 * Input:	n > 2
 *		l >= 0
 *		u any
 * Output:	r1, r2, r3 Results
 */
void factor_brute3(verylong m, ulong l, ulong u,
	IntList& r1, IntList& r2, IntList& r3)
{
	assert(zscompare(m, 3) >= 0);
	assert(u <= MaxUpperBound);

	int		c;
	int		t = 0;
	static verylong	s = 0;
	ulong		s1;
	ulong		f;
	static verylong	m1 = 0;
	static verylong	m2 = 0;
	static verylong	m3 = 0;
	static verylong	m4 = 0;
	static verylong	temp = 0;

	zssub(m, 1, &m1);
	zcopy(m, &m2);
	zsadd(m, 1, &m3);
	zcopy(m3, &m4);			// m4 must be unchanged

	if (u >= l)
	{
		ulong lower = l;
		l = 2;
		while (l < lower)
			l += primeDiff[t++];
	}

	zsqrt(*max3(&m1, &m2, &m3), &s, &temp);	// Upper bound of trial divisions
	if (zscompare(s, u) > 0)
		s1 = u;			// Stop trial divisions at u
	else
		s1 = ztouint(s);

	while (l <= s1)
	{
		f = zsmod(m4, l);
		if (f < 3)
		{
			c = 0;		// Multiplicity of a factor
			if (f == 2 || l == 2 && f == 0)
			{
				while (zsmod(m1, l) == 0)
				{
					zsdiv(m1, l, &m1);
					if (trace_level > 6)
						cout << "m1 = " << m1 << endl;
					c++;
				}
				if (c > 0)
				{
					r1.Add(l, c, NumPrime);
					c = 0;		// In case l == 2
				}
			}
			if (f == 1)
			{
				while (zsmod(m2, l) == 0)
				{
					zsdiv(m2, l, &m2);
					if (trace_level > 6)
						cout << "m2 = " << m2 << endl;
					c++;
				}
				if (c > 0)
					r2.Add(l, c, NumPrime);
			}
			if (f == 0)
			{
				while (zsmod(m3, l) == 0)
				{
					zsdiv(m3, l, &m3);
					if (trace_level > 6)
						cout << "m3 = " << m3 << endl;
					c++;
				}
				if (c > 0)
					r3.Add(l, c, NumPrime);
			}

			if (c > 0)
			{
				zsqrt(*max3(&m1, &m2, &m3), &s, &temp);	// Upper bound of trial divisions
				if (zscompare(s, u) > 0)
					s1 = u;			// Stop trial divisions at u
				else
					s1 = ztouint(s);
			}
		}

		l += primeDiff[t++];
	}

	add_cofactor(m4, r3, l);
	zsdec(m4, 1);
	add_cofactor(m4, r2, l);
	zsdec(m4, 1);
	add_cofactor(m4, r1, l);

} // factor_brute3





/*
 * Prints state in clear text.
 */
void print_state(ushort state)
{
	switch (state)
	{
		case NumPrime:
			cout << "prime";
		break;

		case NumProbPrime:
			cout << "probable prime";
		break;

		case NumComposite:
			cout << "composite";
		break;

		default:
			cout << "Illegal state";
	}

} // print_state





/*
 * Initializes the module.
 */
bool InitBase()
{
	extern void InitArray();

	cout.setf(ios::unitbuf);	// Unbuffered output
	zstart();
	create_incr_list();
	InitArray();

	int fd = open("primediff.dat", O_RDONLY);
	struct stat statbuf;

	if (fstat(fd, &statbuf) == 0 && !S_ISDIR(statbuf.st_mode))
	{
		primeDiff = (ulong *) mmap(NULL, statbuf.st_size, PROT_READ,
			MAP_PRIVATE, fd, 0);
		if (primeDiff == (ulong *) -1)
		{
			cout << "Can't map primediff.dat" << endl;
			primeDiff = NULL;
		}
		else
		{
			if (trace_level > 2)
			{
				// Test
				cout << "The first 13 primes are:" << endl;
				ulong n = 2;
				int i = 0;
				while (i < 13)
				{
					cout << " " << n;
					n += primeDiff[i++];
				}
				cout << endl;
			}
		}
	}
	else
	{
		perror("InitBase failed");
	}

	return primeDiff != NULL;

} // InitBase





/*
 * sum of divisors of n
 * Input: n > 1
 * *r may be n
 * Gives negative result if n cannot be completely factorized.
 */
void factorsum(verylong n, ulong upper, verylong* r, IntList& l)
{
	assert(zscompare(n, 1) > 0);

	static verylong	temp = 0;
	static verylong	temp1 = 0;
	ushort		state;
	ulong		lower = 0;

	state = factor(n, lower, upper, l);
	zone(&temp1);

	if (trace_level > 1)
	{
		cout << "factor returns status: ";
		print_state(state);
		cout << endl;
		cout << l << endl;
	}

	if (state == NumPrime)
	{
		POSITION pos = l.GetHeadPosition();
		while (pos != NULL)
		{
			IntElem* p = l.GetNext(pos);
			powersum(*p->GetNum(), p->GetCount(), &temp);
			zmulin(temp, &temp1);
		}
	}

	zsub(temp1, n, r);

} // factorsum





/*
 * Sieb des Eratosthenes
 * siebt, beginnend bei n, count zahlen unter verwendung der globalen liste
 * inctab. Liefert das ergebnis des
 * siebens, nachdem alle faktoren <= upper ausgesiebt sind.
 * ausgabe ist eine integerzahl, deren binaerdarstellung das sieb
 * darstellt (listen waeren viel langsamer).
 * lsb der ausgabe entspricht erster ungerader zahl des inter-
 * valls [n, n+count-1], naechstes bit der folgenden ungeraden zahl usw. ist
 * das bit 0, ist die entsprechende zahl prim, sonst zusammengesetzt.
 * ab kann auch ungerade sein, n > sqrt (n+count).
 */
void e_sieve(verylong n, verylong count, verylong upper, verylong* r)
{
	assert(zislong(upper));

	ulong		s1;
	static verylong	s = 0;
	static verylong	temp = 0;
	static verylong	temp1 = 0;
	static verylong	loops = 0;
	static verylong	first = 0;
	static verylong	last = 0;
	int		t = 1;			// Index into incr_list[]
	ulong		d = 3;
	ulong		h = d << 1;

	zzero(r);
	zadd(n, count, &last);
	zssub(last, 1, &temp);
	zsqrt(temp, &s, &temp1);
	assert(zcompare(n, s) >= 0);

	if (zcompare(s, upper) > 0)
		s1 = ztouint(upper);
	else
		s1 = ztouint(s);

	do
	{
		// Formula: loops = (count + d) / h;
		zsadd(count, d, &temp);
		zsdiv(temp, h, &loops);		// loops can become 0
		// Formula: first = d + h * ((n + d - 1) / h);
		zsadd(n, d - 1, &temp);
		zsdiv(temp, h, &temp1);
		zsmul(temp1, h, &temp);
		zsadd(temp, d, &first);

		if (zcompare(first, last) < 0)
		{
			// Formula: index = (ulong) ((first - n) / 2)
			zsub(first, n, &temp);
			z2div(temp, &temp);
			ulong index = ztouint(temp);

			do
			{
				zsetbit(r, index);
				index += d;
				zsdec(loops, 1);
			} while (zsign(loops) >= 0);
		}

		if (t >= incr_len)
			t = 5;			// New cycle
		d += incr_list[t++];
		h = d << 1;
	} while (d <= s1);

} // e_sieve





void print_sieve(verylong n, verylong sieve)
{
	verylong n1 = 0;
	verylong sieve1 = 0;

	zcopy(n, &n1);
	zcopy(sieve, &sieve1);

	if (zeven(n1))
		zsinc(n1, 1);

	while (zsign(sieve1) != 0)
	{
		if (zeven(sieve1))
		{
			cout << n1 << " ";
		}

		zsinc(n1, 2);
		zrshift(sieve1, 1, &sieve1);
	}

	cout << endl;

	zfree(&sieve1);
	zfree(&n1);

} // print_sieve





/*
 * Fakultaet
 * Diese Implementierung optimiert die Geschwindigkeit durch Vermeidung
 * von Integer*long-Multiplikationen.
 * Eingabe:	n >= 0
 * Ausgabe:	n!
 */
void fac(ulong n, verylong* r)
{
	zone(r);

	while (n > 0xffff)
	{
		zsmul(*r, n, r);
		--n;
	}

	while (n > 1625)
	{
		zsmul(*r, n * (n - 1), r);
		n -= 2;
	}

	while (n > 255)
	{
		zsmul(*r, n * (n - 1) * (n - 2), r);
		n -= 3;
	}

	while (n > 4)
	{
		zsmul(*r, n * (n - 1) * (n - 2) * (n - 3), r);
		n -= 4;
	}

	if (n > 1)
	{
		ulong f = n;
		if (n > 3) f *= (n - 2);
		if (n > 2) f *= (n - 1);
		zsmul(*r, f, r);
	}

} // fac





/*
 *
 */
ulong gcd(ulong a, ulong b)
{
	ulong r;

	while (b > 0)
	{
		r = a % b;
		a = b;
		b = r;
	}

	return a;

} // gcd





/*
 * Berechnung des Binominalkoeffizienten n ueber k.
 * Eingabe n >= k >= 0
 */
void binomial(ulong n, ulong k, verylong* r)
{
	ulong	i;

	assert(n >= k);

	zone(r);
	if (2 * k > n)
		k = n - k;

	if (k <= 1)
	{
		if (k != 0)
			zuintoz(n, r);		// Otherwise r = 1
		return;
	}

	ulong		lower = n - k + 1;
	ulong		denominator = 2;
	static verylong	d = 0;
	static verylong	g = 0;

	zuintoz(denominator, &d);

	for (i = 0; i < k; i++)
	{
		zsmul(*r, lower + i, r);
		if (trace_level > 4)
			cout << "r = " << *r << endl;
		zgcd(*r, d, &g);
		if (zscompare(g, 1) > 0)
		{
			if (trace_level > 4)
				cout << "Divide by " << g << endl;
			ulong gl = ztouint(g);
			zsdiv(*r, gl, r);
			zsdiv(d, gl, &d);
			if (zscompare(d, 1) == 0 && denominator < k)
				zsadd(d, ++denominator, &d);
		}
	}

	if (zscompare(d, 1) > 0)
	{
		if (trace_level > 4)
			cout << "Divide by " << d << endl;
		zdiv(*r, d, r, &g);
	}

	for (i = ++denominator; i <= k; i++)
	{
		if (trace_level > 4)
			cout << "Divide by " << i << endl;
		zsdiv(*r, i, r);
	}

} // binomial
