#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "base.h"

/* Simple interface to the library */

const size_t InputLength = 1024;

static int nextindex;
static bool tokenSeen;

static char* Function[] =
{
	"powersum",
	"factorsum",
	"fac",
	"binomial",
	"expt",
	"expt_mod",
	"e_sieve"
};

static const int NumFunctions = sizeof Function / sizeof (char *);

enum FunctionList
{
	F_POWERSUM, F_FACTORSUM,
	F_FAC, F_BINOMIAL,
	F_EXPT, F_EXPT_MOD,
	F_E_SIEVE
};

static int FuncArgCount[] =
{
	2, 1, 1, 2, 2, 3, 3
};

static bool GetExpression(const char* s, verylong* result);

static void Help()
{
	cerr << "Commands" << endl;
	cerr << "========" << endl;
	cerr << "q, quit, exit\tleave program" << endl;
	cerr << "help, ?\t\tshow help" << endl;
	cerr << "trace <n>\tset trace_level to <n>" << endl;
	cerr << "<expression>\tcalculate <expression> and print result" << endl;
	cerr << "" << endl;
} // Help

static bool ExecFunction(int function, verylong* argument, verylong* result)
{
	switch (function)
	{
		case F_POWERSUM: // powersum
			if (zsign(argument[1]) == 0)
			{
				cerr << "Argument 2 of powersum() negative" << endl;
				return false;
			}
			if (zislong(argument[1]) == false)
			{
				cerr << "Argument 2 of powersum() too big" << endl;
				return false;
			}
			powersum(argument[0], ztouint(argument[1]), result);
		break;

		case F_FACTORSUM: // factorsum
		{
			IntList l;
			factorsum(argument[0], 1000, result, l);
		}
		break;

		case F_FAC: // fac
			if (zislong(argument[0]) == false)
			{
				cerr << "Argument of fac() too big" << endl;
				return false;
			}
			fac(ztouint(argument[0]), result);
		break;

		case F_BINOMIAL: // binomial
			if (zislong(argument[0]) == false)
			{
				cerr << "Argument 1 of binomial() too big" << endl;
				return false;
			}
			if (zislong(argument[1]) == false)
			{
				cerr << "Argument 2 of binomial() too big" << endl;
				return false;
			}
			if (zcompare(argument[0], argument[1]) < 0)
			{
				cerr << "Argument 2 of binomial() must be smaller than argument 1" << endl;
				return false;
			}
			binomial(ztoint(argument[0]), ztoint(argument[1]), result);
		break;

		case F_EXPT: // zexp
			zexp(argument[0],
				argument[1], result);
		break;

		case F_EXPT_MOD: // zexpmod
			zexpmod(argument[0],
				argument[1], argument[2], result);
		break;

		case F_E_SIEVE: // e_sieve
			e_sieve(argument[0],
				argument[1], argument[2], result);
		break;

		default:
			cerr << "Internal error, ExecFunction("
				<< function << ")" << endl;
			return false;
	} // switch

	return true;

} // ExecFunction

static bool SkipBlanks(const char* s)
{
	while (isspace(s[nextindex]))
		nextindex++;

	if (s[nextindex] == '\0')
		return false;

	return true;

} // SkipBlanks

static bool GetExprList(const char* s, verylong* result, int& count)
{
	count = 0;

	tokenSeen = false;
	if (GetExpression(s, &result[count]) == false)
	{
		tokenSeen = true;
		return false;
	}

	if (tokenSeen == false)
	{
		cerr << "Empty expression in ()" << endl;
		tokenSeen = true;
		return false;
	}

	do
	{
		if (SkipBlanks(s) == false)
			return true;

		count++;
		if (s[nextindex] == ',')
		{
			nextindex++;
			tokenSeen = false;

			if (GetExpression(s, &result[count]) == false)
			{
				tokenSeen = true;
				return false;
			}

			if (tokenSeen == false)
			{
				cerr << "Empty expression in ()" << endl;
				tokenSeen = true;
				return false;
			}
		}
		else
		{
			return true;
		}
	} while (1);

	return true;	// Never reached

} // GetExprList

static bool GetSimpleAux(const char* s, const char* func, int funcArgCount,
	verylong* arguments, int& count)
{
	if (GetExprList(s, arguments, count) == false)
		return false;

	if (SkipBlanks(s) == false)
	{
		cerr << "Missing ')' at end of expression" << endl;
		return false;
	}

	if (s[nextindex] != ')')
	{
		cerr << "Missing ')'" << endl;
		return false;
	}

	nextindex++;
#if 0
	cout << Function[i] << "(";
	int j;
	for (j = 0; j < count - 1; j++)
		cout << arguments[j] << ", ";
	if (count > 0)
		cout << arguments[j];
	cout << ")" << endl;
#endif
	if (count != funcArgCount)
	{
		cerr << "Function " << func << " must have "
			<< funcArgCount << " arguments" << endl;
		return false;
	}

	return true;

} // GetSimpleAux

static bool GetSimple(const char* s, verylong* result)
{
	if (SkipBlanks(s) == false)
		return false;

	tokenSeen = true;

	if (s[nextindex] == '(')
	{
		nextindex++;
		tokenSeen = false;

		if (GetExpression(s, result) == false)
		{
			tokenSeen = true;
			return false;
		}

		if (tokenSeen == false)
		{
			cerr << "Empty expression in ()" << endl;
			tokenSeen = true;
			return false;
		}

		if (SkipBlanks(s) == false)
		{
			cerr << "Incomplete expression" << endl;
			return false;
		}

		if (s[nextindex] != ')')
		{
			cerr << "Missing ')'" << endl;
			return false;
		}

		nextindex++;
		return true;
	}

	if (isdigit(s[nextindex]) || s[nextindex] == '-' || s[nextindex] == '+')
	{
		char p[InputLength];
		int oldindex = nextindex;
		do
		{
			p[nextindex - oldindex] = s[nextindex];
			nextindex++;
		} while (isdigit(s[nextindex]));
		p[nextindex - oldindex] = '\0';
		zsread(p, result);
		return true;
	}

	if (isalpha(s[nextindex]))
	{
		char func[InputLength];
		int oldindex = nextindex;

		while (isalpha(s[nextindex]) || s[nextindex] == '_')
			nextindex++;

		memcpy(func, &s[oldindex], nextindex - oldindex);
		func[nextindex - oldindex] = '\0';

		int i;
		for (i = 0; i < NumFunctions; i++)
		{
			if (nextindex - oldindex == strlen(Function[i])
				&& memcmp(func, Function[i], nextindex - oldindex) == 0)
				break;
		}

		if (i == NumFunctions)
		{
			cerr << "Unknown function " << func << endl;
			return false;
		}

		if (SkipBlanks(s) == false)
		{
			cerr << "Missing parameters for function "
				<< func << endl;
			return false;
		}

		if (s[nextindex] != '(')
		{
			cerr << "Expected '(' after function "
				<< func << endl;
			return false;
		}
		nextindex++;

		int count;
		verylong arguments[128];

		memset(arguments, 0, sizeof arguments);

		bool retval = false;
		if (GetSimpleAux(s, Function[i], FuncArgCount[i], arguments, count))
			retval = ExecFunction(i, arguments, result);

		for (i = 0; i < 128; i++)
			zfree(&arguments[i]);
		return retval;
	}

	cerr << "Missing number, function, or '('" << endl;
	return false;

} // GetSimple

static bool GetFactor(const char* s, verylong* result)
{
	if (GetSimple(s, result) == false)
		return false;

	do
	{
		if (SkipBlanks(s) == false)
			return true;

		verylong temp = 0;

		if (s[nextindex] == '*')
		{
			nextindex++;
			if (GetSimple(s, &temp) == false)
			{
				zfree(&temp);
				return false;
			}

			zmulin(temp, result);
		}
		else if (s[nextindex] == '/')
		{
			nextindex++;
			if (GetSimple(s, &temp) == false)
			{
				zfree(&temp);
				return false;
			}

			if (zsign(temp) == 0)
			{
				cerr << "Division by 0" << endl;
				zfree(&temp);
				return false;
			}

			zdiv(*result, temp, result, &temp);
		}
		else
		{
			return true;
		}
		zfree(&temp);
	} while (1);

	return true;	// Never reached

} // GetFactor

static bool GetExpression(const char* s, verylong* result)
{
	if (GetFactor(s, result) == false)
		return false;

	do
	{
		verylong temp = 0;

		if (SkipBlanks(s) == false)
			return true;

		if (s[nextindex] == '+')
		{
			nextindex++;
			if (GetFactor(s, &temp) == false)
			{
				zfree(&temp);
				return false;
			}

			zadd(*result, temp, result);
		}
		else if (s[nextindex] == '-')
		{
			nextindex++;
			if (GetFactor(s, &temp) == false)
			{
				zfree(&temp);
				return false;
			}

			zsub(*result, temp, result);
		}
		else
		{
			return true;
		}
		zfree(&temp);
	} while (1);

	return true;	// Never reached

} // GetExpression

int main(int argc, char** argv)
{
	char s[InputLength];

	trace_level = 0;
	InitBase();

	bool interactive = (argc <= 1);

	if (!interactive)
	{
		if (strlen(argv[1]) >= InputLength)
		{
			cout << "calc: command line too long" << endl;
			return 1;
		}
		strcpy(s, argv[1]);
	}

	do
	{
		if (interactive)
		{
			cout << "calculate: " << flush;
			cin.getline(s, sizeof s);
			if (cin.eof())
				break;

			if (!cin)
			{
				cerr << "Input line too long" << endl;
				continue;
			}
		}

		if (strcmp(s, "?") == 0 || strcmp(s, "help") == 0)
		{
			Help();
		}
		else if (strcmp(s, "exit") == 0
			|| strcmp(s, "q") == 0 || strcmp(s, "quit") == 0)
		{
			break;
		}
		else if (memcmp(s, "trace", 5) == 0)
		{
			int n;

			if (sscanf(s, "trace %d", &n) == 1)
				trace_level = n;
			else
				cerr << "use \"trace <level>\"" << endl;
		}
		else
		{
			static verylong result = 0;

			nextindex = 0;
			tokenSeen = false;
			if (GetExpression(s, &result))
			{
				if (s[nextindex] == '\0')
					cout << result << endl;
				else
					;//cout << "Syntax error" << endl;
			}
			else
			{
				//if (tokenSeen)
					//cout << "Syntax error" << endl;
			}
		}
	} while (interactive);

	return 0;
}

