#include <assert.h>
#include "base.h"


/*
 * Factorization with a Monte Carlo method (Pollard Rho?).
 * Only step 1 implemented.
 * Input:	n > 0	Number to factorize
 *		lim	Maximum in thousands of iterations
 * Output:	Factor or 1
 * Return:	NumProbPrime if a factor > 1 has been found;
 *		NumComposite otherwise
 */
ushort mc_factor(verylong n, verylong lim, verylong* factor)
{
	static verylong	xslow = 0, xsold = 0;
	static verylong	xfast = 0, xfold = 0;
	static verylong	q = 0, qold = 0;
	static verylong	count = 0;
	static verylong	limit = 0;
	static verylong	temp = 0;
	ulong		inc = 10000;
	ulong		c = inc + 1;		// Loop counter

	assert(zsign(n) > 0);

	if (trace_level > 0)
		cout << "mc_factor: n = " << n << endl;

	zuintoz(2, &xslow); zuintoz(3, &xfast);
	zuintoz(2, &xsold); zuintoz(3, &xfold);
	zone(&q); zone(&qold);
	zone(factor);

	zsmul(lim, inc, &limit);

	while (zcompare(count, limit) < 0)
	{
		// Formula: xslow = (xslow * xslow - 1) % n;
		zsqinmod(&xslow, n);
		zsadd(xslow, -1, &xslow);
		// Formula: xfast = (xfast * xfast) % n;
		zsqinmod(&xfast, n);
		// Formula: xfast = (xfast * (xfast - 2)) % n;
		// Note: x*(x-2) = (x-1)^2-1
		zsadd(xfast, -1, &xfast);
		zsqinmod(&xfast, n);
		zsadd(xfast, -1, &xfast);
		zsub(xfast, xslow, &temp);
		if (zsign(temp) != 0)
		{
			// Formula: q = (q * (xfast - xslow)) % n;
			zmulinmod(temp, &q, n);
		}

		if (--c == 0)
		{
			zgcd(q, n, factor);
			if (trace_level > 1)
			{
				cout << "  Count = " << count;
				cout << "  gcd = " << *factor << endl;
			}
			if (trace_level > 5)
			{
				cout << "  q = " << q;
				cout << "  xslow = " << xslow;
				cout << "  xfast = " << xfast << endl;
			}
			if (zcompare(*factor, n) == 0)
			{
				zcopy(xsold, &xslow); zcopy (xfold, &xfast);
				zcopy(qold, &q);
				zone(factor);
				inc /= 10;
				if (trace_level > 1)
					cout << "  Retry with increment " << inc << endl;
			}
			zcopy(xslow, &xsold); zcopy (xfast, &xfold);
			zcopy(q, &qold);
			if (zscompare(*factor, 1) > 0)
				return NumProbPrime;
			c = inc;
		}

		zsadd(count, 1, &count);
	}

	return NumComposite;

} // mc_factor





// Handle a factor f of n.
// r must be initialized correctly.
ushort handle_factor(verylong n, verylong f, ulong l, ulong u, IntList& r)
{
	verylong	cofactor = 0;	// Not static, recursion!
	verylong	f1 = 0;		// Not static, recursion!
	static verylong	temp = 0;
	static verylong	temp1 = 0;
	IntList		r1;
	IntList		r2;
	int		mult = 1;
	ushort		s;

	if (trace_level > 0)
	{
		cout << "handle_factor: n = " << n;
		cout << ", factor = " << f << endl;
	}

	zcopy(f, &f1);

	if (zscompare(f1, 1) == 0 || zcompare(f1, n) == 0)
	{
		cout << "handle_factor: trivial factor 1" << endl;
		zone(&f1);
	}

	zdiv(n, f1, &cofactor, &temp);		// cofactor = n / f1
	assert(zsign(temp) == 0);

	if (zscompare(f1, 1) != 0)
	{
		do
		{
			zdiv(cofactor, f1, &temp, &temp1);
			if (zsign(temp1) == 0)			// Rest 0?
			{
				zcopy(temp, &cofactor);		// Quotient
				mult++;
			}
		} while (zsign(temp1) == 0);

		if (zscompare(cofactor, 1) == 0)
		{
			zcopy(f1, &cofactor);
		}
		else
		{
			// Factorize f1.
			s = factor(f1, l, u, r1);
			assert(s == NumPrime);
		}

	}

	// Factorize cofactor.
	s = factor(cofactor, l, u, r2);
	assert(s == NumPrime);

	if (mult > 1)
	{
		// Multiply all factor counts.
		POSITION pos = r2.GetHeadPosition();
		while (pos != NULL)
		{
			IntElem* p = r2.GetNext(pos);
			p->SetCount(mult * p->GetCount());
		}
	}

	// Sort r2, r1 into r.
	r.Merge(r1, r2);

	zfree(&f1);
	zfree(&cofactor);

	if (trace_level > 1)
		cout << "handle_factor(" << n << ") completed" << endl;

	return NumPrime;

} // handle_factor





ushort complete(IntList& r, ulong l, ulong u)
{
	if (trace_level > 1)
		cout << "  Completing factorization..." << endl;

	assert(r.GetState() == NumProbPrime);

	verylong	f = 0;		// Not static, recursion!
	verylong	num = 0;

	zcopy(*r.GetNum(), &num);
	ushort s = prove_prime(num, l, u, &f);

	if (s == NumComposite && zscompare(f, 1) > 0 && zcompare(f, num) < 0)
	{
		// Nichttrivialen Faktor gefunden.
		r.Remove();
		s = handle_factor(num, f, l, u, r);
	}
	else if (s == NumPrime)
	{
		ushort count = r.GetCount();
		r.Remove();
		r.Add(num, count, NumPrime);
	}

	zfree(&num);
	zfree(&f);

	return s;
																																							 
} // complete





/*
 * Attempt to find at least one factor of num (which is a factor of n) using
 * various methods.
 * Input:	n	The original number to factorize
 *		num	The number to factorize
 *		u	num is known to have no factor up to u
 *		r	The list of factors of n found so far
 * Output:	num	num after removal of small factors
 *		f	A factor (may be a trivial factor)
 *		r	factor_brute may add factors
 * Return:	NumComposite if no factor has been found;
 *		NumProbPrime if after trial division the state of the remainder
 *		could not be determined (should not happen); NumPrime otherwise
 */
ushort factor_aux(verylong n, verylong* num, ulong u, verylong* f, IntList& r)
{
	static verylong	temp = 0;
	ushort		s;

	s = pollard_factor(*num, 10000, 10000, f);
	if (s == NumComposite)
		s = pplus1_factor(*num, 10000, 10000, f);

	// Up to here, we were quick. Now it gets time consuming...
	if (s == NumComposite && u < MaxUpperBound)
	{
		r.Remove();
		factor_brute(n, u, MaxUpperBound, r);
		//assert(zscompare(*f, 1) == 0 || zcompare(*f, n) == 0);
		if (zscompare(*f, 1) != 0 && zcompare(*f, n) != 0)
		{
			cout << "ERROR: f should be 1, is " << *f << endl;
			cout << "in factorization of " << n << endl;
		}
		s = r.GetState();
		if (s == NumProbPrime)
		{
			if (zcompare(*r.GetNum(), *num) == 0)
			{
				// No additional factor found.
				s = NumComposite;
			}
			else
			{
				s = complete(r, MaxUpperBound, MaxUpperBound);
				// This assert could fail with *really* big
				// numbers. Have to find a solution.
				assert(s == NumPrime);
				zcopy(*r.GetNum(), num);
			}
		}
		else if (s == NumComposite)
		{
			assert(!r.IsEmpty());
			zcopy(*r.GetNum(), num);
		}
	}

	if (s == NumComposite)
	{
		if (trace_level > 0)
			cout << "Trying ECM... " << endl;
		s = zecm(*num, f, 0, 100, 125, 5, 0, trace_level)
			== 1 ? NumPrime : NumComposite;
		if (trace_level > 0)
			cout << "Finished with ECM" << endl;
	}

	if (s == NumComposite)
	{
		// limit can be static, even if we recurse,
		// because it is constant.
		static verylong limit = 0;
		zuintoz(200, &limit);
		mc_factor(*num, limit, f);
		if (zscompare(*f, 1) > 0 && zcompare(*f, *num) < 0)
			s = NumPrime;
	}

	if (s == NumComposite)
	{
		if (trace_level > 0)
			cout << "Trying zpollardrho... " << endl;
		s = zpollardrho(*num, f, &temp, 1000000)
			> 0 ? NumPrime : NumComposite;
		if (trace_level > 0)
			cout << "Finished with zpollardrho" << endl;
	}

	if (s == NumComposite)
	{
		if (trace_level > 0)
			cout << "Trying ECM... " << endl;
		s = zecm(*num, f, 0, 200, 6250, 5, 0, trace_level)
			== 1 ? NumPrime : NumComposite;
		if (trace_level > 0)
			cout << "Finished with ECM" << endl;
	}

	if (s == NumComposite)
		s = pollard_factor(*num, 1000000, MaxUpperBound, f);

	if (s == NumComposite)
		s = pplus1_factor(*num, 1000000, MaxUpperBound, f);

	return s;

} // factor_aux





/*
 * Completely factorize n.
 * Input:	n	> 0
 * Output:	r	factor list
 * Return:	Status of the greatest factor of n. If this
 *		is not NumPrime the factorization could not be completed.
 * Bugs:	1	is returned as "prime"
 */
ushort factor(verylong n, ulong l, ulong u, IntList& r)
{
	verylong	f = 0;
	ushort		s;

	if (trace_level > 0)
		cout << "factor: n = " << n << endl;

	assert(zsign(n) > 0);
	assert(u >= l);

	if (zscompare(n, 2) > 0)
		factor_brute(n, l, u, r);
	else
		r.Add(n, 1, NumPrime);

	if (trace_level > 1)
		cout << "Factors found until now:" << r << endl;

	// The indivisible remainder, if existent, has no more factors up to u.
	// If it is a probable prime, try to prove it prime.
	s = r.GetState();
	if (s == NumProbPrime)
		s = complete(r, l, u);

	if (s == NumComposite)
	{
		verylong num = 0;
		zcopy(*r.GetNum(), &num);
		s = factor_aux(n, &num, u, &f, r);
		if (zscompare(f, 1) > 0 && zcompare(f, num) < 0)
		{
			r.Remove();
			s = handle_factor(num, f, l, u, r);
		}
		zfree(&num);
	}

	zfree(&f);

	if (trace_level > 1)
	{
		cout << "factor result: ";
		print_state(s);
		cout << endl;
	}

	return s;

} // factor
