#include <assert.h>
#include "base.h"

const ulong DIFNPMAX = 100;


/* Performs the standard 2nd phase */
/* p is the least prime greater than b1 */
bool pollardstep2(verylong n, verylong a, ulong p, int t,
	ulong b2, verylong* factor)
{
	static verylong	a1 = 0;
	static verylong	prod = 0;
	static verylong	temp = 0;
	static verylong	tabpower[DIFNPMAX];
	static bool	tabInitialized = false;
	int		nbp = 0;
	ulong		nbgcd = b2 / 100;
	ulong		pDiff;

	if (trace_level > 1)
		cout << "pollardstep2: n = " << n
			<< ", b2 = " << b2 << endl;

	if (nbgcd == 0)
		nbgcd = 1;

	if (tabInitialized == false)
	{
		tabpower[0] = 0;
		zsqmod(a, n, &tabpower[0]);
		for (unsigned i = 1; i < DIFNPMAX; i++)
		{
			// Formula: tabpower[i] = (tabpower[i - 1] * tabpower[0]) % n;
			tabpower[i] = 0;
			zmulmod(tabpower[i - 1], tabpower[0], n, &tabpower[i]);
		}
		tabInitialized = true;
	}

	zcopy(a, &a1);			// Original a may not be changed
	zsexpmod(a1, p, n, &a1);	// a1 = a1**p mod n
	zssub(a1, 1, &prod);		// prod = a1 - 1

	do
	{
		pDiff = primeDiff[t - 1] / 2;
		assert(pDiff < DIFNPMAX);
		// Formula: a1 = (a1 * tabpower[pDiff]) % n;
		zmulmod(a1, tabpower[pDiff], n, &a1);
		assert(a1 != 0);
		// Formula: prod = (prod * (a1 - 1)) % n;
		zssub(a1, 1, &temp);
		zmulmod(prod, temp, n, &prod);

		if(++nbp % nbgcd == 0)
		{
			zgcd(prod, n, factor);
			if (zscompare(*factor, 1) > 0)
				return true;
		}

		p += primeDiff[t++];
	} while (p <= b2);

	if (nbp % nbgcd)
	{
		if (trace_level > 1)
			cout << "  Performing last gcd in step 2" << endl;
		zgcd(prod, n, factor);
		if (zscompare(*factor, 1) > 0)
			return true;
	}

	return false;

} // pollardstep2





ushort pollard_factor(verylong n,
	ulong b1, ulong b2, verylong* factor)
{
	// The value of a is no longer needed after recursion.
	static verylong	a = 0;

	static verylong	temp = 0;

	bool		done = false;
	int		t = 1;
	ulong		nbp = 0;
	ulong		p = 3;		// Current divisor
	ulong		q;
	ulong		m;
	ulong		nbgcd = b1 / 100;

	if (trace_level > 0)
		cout << "Pollard P-1: N = " << n
			<< ", b1 = " << b1 << ", b2 = " << b2 << endl;

	if (nbgcd == 0)
		nbgcd = 1;

	zuintoz(2, &a);

	// First handle base 2.
	q = 2;
	m = b1 / q;

	while (q < m)
		q <<= 1;

	// Calculate 2^q mod n
	zsexpmod(a, q, n, &a);

	// If the remainder is 1, the method fails.
	if (zscompare(a, 1) == 0)
		return NumProbPrime;

	zone(factor);

	do
	{
		nbp++;
		m = b1 / p;
		q = p;
		while (q < m)
			q *= p;
		zsexpmod(a, q, n, &a);
		if (nbp % nbgcd == 0)
		{
			zssub(a, 1, &temp);
			zgcd(temp, n, factor);
			if (zscompare(*factor, 1) > 0)
			{
				if (trace_level > 1)
					cout << "  Step 1 found factor " << *factor
						<< " at p = " << p << endl;
				if (p > 3 && zcompare(*factor, n) == 0)
				{
					p = 63 * p / 64;
					pollard_factor(n, p, b2, factor);
					// The resulting factor can be 1. Shit happens.
				}
				done = true;
				nbp = nbgcd;	// Prevent duplicate test
				break;
			}
		}
		if (zscompare(a, 1) == 0)
			break;			// The next block will handle it
		p += primeDiff[t++];
	} while (p <= b1);

	if (nbp % nbgcd)
	{
		if (trace_level > 1)
			cout << "  Performing last gcd in step 1" << endl;
		zssub(a, 1, &temp);
		zgcd(temp, n, factor);
		if (zscompare(*factor, 1) > 0)
		{
			if (trace_level > 1)
				cout << "  Step 1 found factor " << *factor
					<< " at p = " << p << endl;
			if (zcompare(*factor, n) == 0)
			{
				pollard_factor(n, b1 / 2, b2, factor);
				// The resulting factor can be 1. Shit happens.
			}
			done = true;
		}
	}

	while (!done)
	{
		done = pollardstep2(n, a, p, t, b2, factor);
		if (done && trace_level > 4)
			cout << "pollardstep2 found factor " << *factor << endl;

		if (done && zcompare(*factor, n) == 0)
		{
			// It would be more efficient to loop within step2.
			done = false;
			b2 /= 2;
			if (b2 < 10)
				break;		// Not successful
		}
		else
		{
			break;
		}
	}

	if (done && zscompare(*factor, 1) > 0 && zcompare(*factor, n) < 0)
		return NumPrime;

	return NumComposite;

} // pollard_factor

