#include <assert.h>
#include "base.h"


extern ushort complete(IntList& r, ulong l, ulong u);
extern ushort factor_aux(verylong n, verylong* num, ulong u, verylong* f,
	IntList& r);


static ushort find_factor(verylong n, ulong u, verylong* factor)
{
	IntList		l;
	static verylong	num = 0;

	zcopy(n, &num);
	l.Add(num, 1, NumComposite);
	ushort s = factor_aux(n, &num, u, factor, l);
	if (s == NumComposite)
	{
		return NumProbPrime;	// No factor found
	}
	else
	{
		if (zscompare(*factor, 1) == 0 || zcompare(*factor, n) == 0)
		{
			if (trace_level > 0)
				cout << "Copying factor from factor list" << endl;
			zcopy(*l.GetNum(), factor);
		}
		if (trace_level > 0)
			cout << "find_factor finds factor " << *factor << endl;
		return NumComposite;
	}

} // find_factor





/*
 * Primality test with theorem 8.
 * Input:	n > 1	Number to prove prime
 *		r	Factorization of n + 1
 *		state	Status of the greatest factor of n + 1
 *		u	n is known to have no factors up to u
 * Output:	factor	A factor (may be trivial) of n if n is found composite
 * Return:	Status of n
 */
ushort theorem8(verylong n, const IntList& r,
	ushort state, ulong upper, verylong* factor)
{
	assert(zscompare(n, 1) > 0);

	static verylong	m = 0;
	int		p = 2;
	int		q = 1;
	int		t = 0;

	(void) state;
	(void) upper;

	zone(factor);
	zsadd(n, 1, &m);		// m = n + 1

	if (trace_level > 1)
		cout << "theorem8: n = " << n << endl;

	static verylong	u = 0, v = 0;
	static verylong	temp = 0;

	while (t < 500)
	{
		if (trace_level > 4)
			cout << "  Trying p = " << p << ", q = " << q << endl;

		lucas(m, n, p, q, &u, &v);
		if (trace_level > 6)
			cout << "  u = " << u << endl;

		if (zsign(u) == 0)
		{
			POSITION pos = r.GetHeadPosition();
			while (pos != NULL)
			{
				IntElem* e = r.GetNext(pos);
				if (trace_level > 1)
					cout << "    Test factor " << *e->GetNum() << endl;

				zdiv(m, *e->GetNum(), &temp, &u);
				lucas(temp, n, p, q, &u, &v);
				if (zsign(u) > 0)
				{
					zgcd(u, n, factor);
					assert(zsign(*factor) > 0);
					if (zscompare(*factor, 1) > 0 && zcompare(*factor, n) < 0)
					{
						if (trace_level > 1)
							cout << "  Found factor: " << *factor << endl;
						return NumComposite;
					}
				}
			}

			assert(pos == NULL);
			if (trace_level > 1)
				cout << "theorem8 result: prime" << endl;
			return NumPrime;
		}
		else
		{
			zgcd(u, n, factor);
			assert(zsign(*factor) > 0);
			if (zscompare(*factor, 1) > 0 && zcompare(*factor, n) < 0)
			{
				if (trace_level > 1)
					cout << "  Found factor: " << *factor << endl;
				return NumComposite;
			}
			p += (int) primeDiff[t++];
		}
	}

	if (trace_level > 1)
		cout << "theorem8: no decision" << endl;
	return NumProbPrime;

} // theorem8





/*
 * Primality test with theorem 9.
 * Input:	n > 1	Number to prove prime
 *		r	Factorization of n - 1
 *		state	Status of the greatest factor of n - 1
 *		u	n is known to have no factors up to u
 * Output:	factor	A factor (may be trivial) of n if n is found composite
 * Return:	Status of n
 */
ushort theorem9(verylong n, const IntList& r,
	ushort state, ulong u, verylong* factor)
{
	assert(zscompare(n, 1) > 0);

	static verylong	m = 0;

	zone(factor);		// *factor = 1
	zsadd(n, -1, &m);

	if (trace_level > 1)
		cout << "theorem9: n = " << n << endl;

	if (state == NumProbPrime || state == NumComposite)
	{
		if (trace_level > 1)
			cout << "  n-1 is not completely factorized, trying theorem10" << endl;
		return theorem10(n, r, state, u, factor);
	}

	assert(state == NumPrime);

	if (trace_level > 3)
	{
		cout << "  u = " << u;
		cout << ", r.GetNum() = " << *r.GetNum() << endl;
		cout << "  factors of " << m << ":";
		cout << r << " (";
		print_state(r.GetState());
		cout << ")" << endl;
	}

	POSITION pos = r.GetHeadPosition();
	while (pos != NULL)
	{
		IntElem* p = r.GetNext(pos);
		static verylong	f = 0;
		static verylong	base = 0;
		static verylong	temp = 0;
		int t = 0;

		zuintoz(2, &base);
		zdiv(m, *p->GetNum(), &f, &temp);

		if (trace_level > 3)
		{
			cout << "  Factor " << *p->GetNum();
			cout << " of " << m << endl;
		}

		while (1)
		{
			if (trace_level > 3)
				cout << "    Test base " << base << endl;

			zexpmod_m_ary(base, f, n, &temp, 1);
			if (zscompare(temp, 1) > 0)
				break;

			if (t > 200)
			{
				// There are numbers for which no base can be
				// found. Are they all Carmichael numbers?
				// Anyway, assume they are composite and try
				// to find a factor.
				// Can't call factor because this would
				// attempt do do another primality proof...
				if (trace_level > 1)
					cout << "  No base found, trying to find a factor..." << endl;
				return find_factor(n, u, factor);
			}

			zsadd(base, primeDiff[t++], &base);
		}

		zexpmod_m_ary(base, m, n, &temp, 1);
		if (zscompare(temp, 1) > 0)
		{
			if (trace_level > 1)
				cout << "theorem9 result: composite" << endl;
			return NumComposite;
		}
	}

	if (trace_level > 1)
		cout << "theorem9 result: prime" << endl;
	return NumPrime;

} // theorem9





/*
 * Primality test with theorem 10.
 * Input:	n > 1	Number to prove prime
 *		r	Factorization of n - 1
 *		state	Status of the greatest factor of n - 1
 *		u	n is known to have no factors up to u
 * Output:	factor	A factor (may be trivial) of n if n is found composite
 * Return:	Status of n
 */
ushort theorem10(verylong n, const IntList& r,
	ushort state, ulong u, verylong* factor)
{
	assert(zscompare(n, 1) > 0);

	static verylong	n1 = 0;
	static verylong	f1 = 0;
	static verylong	f2 = 0;
	static verylong	base = 0;
	static verylong	temp = 0;

	if (trace_level > 0)
		cout << "theorem10: n = " << n << endl;

	// If all factors of n-1 are known theorem9 is faster.
	if (state == NumPrime)
		return theorem9(n, r, state, u, factor);

	zone(factor);
	zsadd(n, -1, &n1);

	// Since the last element of r is the unfactorized part,
	// it can only have a count of 1.
	assert(r.GetCount() == 1);
	zdiv(n1, *r.GetNum(), &f1, &temp);

	if (trace_level > 3)
	{
		cout << "  u = " << u;
		cout << ", r.GetNum() = " << *r.GetNum() << endl;
		cout << "  factors of " << n1 << ":";
		cout << r << " (";
		print_state(r.GetState());
		cout << ")" << endl;
	}

	zsmul(f1, u, &temp);
	zsmul(temp, u, &temp);		// temp = f1 * u^2
	if (zcompare(temp, *r.GetNum()) < 0)
	{
		if (trace_level > 1)
			cout << "  cannot decide, trying theorem11" << endl;
		return theorem11(n, r, state, u, factor);
	}

	int t = 0;
	ztwo(&base);

	while (1)
	{
		if (trace_level > 3)
			cout << "  Test base " << base << endl;

		zexpmod_m_ary(base, n1, n, &temp, 1);
		if (zscompare(temp, 1) > 0)
		{
			if (trace_level > 1)
				cout << "  composite because of Fermat test" << endl;
			return NumComposite;
		}

		zexpmod_m_ary(base, f1, n, &temp, 1);
		zsadd(temp, -1, &temp);
		zgcd(temp, n, factor);
		if (zscompare(*factor, 1) == 0)
		{
			POSITION	pos = r.GetHeadPosition();
			bool		prime = true;

			while (pos != NULL)
			{
				IntElem* p = r.GetNext(pos);
				// The last iteration is unnecessary because at
				// this point n is already proven to be prime.
				if (pos == NULL)
					break;
				zdiv(n1, *p->GetNum(), &f2, &temp);
				zexpmod_m_ary(base, f2, n, &temp, 1);
				zsadd(temp, -1, &temp);
				zgcd(temp, n, factor);
				if (zscompare(*factor, 1) > 0)
				{
					prime = false;
					// base**(n-1)/q-1 could be divisible by n in which case
					// the gcd would become n.
					if (zcompare(*factor, n) < 0)
					{
						if (trace_level > 1)
							cout << "  Found a nontrivial factor: " << *factor << endl;
						return NumComposite;
					}
				}
			}

			if (prime)
			{
				if (trace_level > 1)
					cout << "theorem10 result: prime" << endl;
				return NumPrime;
			}
		}
		else
		{
			// base**f1-1 could be divisible by n in which case
			// the gcd would become n.
			if (zscompare(*factor, 1) > 0 && zcompare(*factor, n) < 0)
			{
				if (trace_level > 1)
					cout << "  Found a nontrivial factor: " << *factor << endl;
				return NumComposite;
			}
		}

		if (t > 200)
		{
			if (trace_level > 1)
				cout << "  No base found, trying to find a factor..." << endl;
			return find_factor(n, u, factor);
		}

		zsadd(base, primeDiff[t++], &base);
	} // while

	assert(false);
	return 0; 		// Not reached

} // theorem10





/*
 * Primality test with theorem 11.
 * Input:	n > 1	Number to prove prime
 *		r	Factorization of n - 1
 *		state	Status of the greatest factor of n - 1
 *		u	n is known to have no factors up to u
 * Output:	factor	A factor (may be trivial) of n if n is found composite
 * Return:	Status of n
 */
ushort theorem11(verylong n, const IntList& r,
	ushort state, ulong u, verylong* factor)
{
	assert(zscompare(n, 1) > 0);

	static verylong	n1 = 0;
	static verylong	f1 = 0;
	static verylong	f2 = 0;
	static verylong	base = 0;
	static verylong	temp = 0;

	if (trace_level > 0)
		cout << "theorem11: n = " << n << endl;

	// If all factors of n-1 are known theorem9 is faster.
	if (state == NumPrime)
		return theorem9(n, r, state, u, factor);

	zone(factor);		// *factor = 1
	zsadd(n, -1, &n1);	// n1 = n - 1

	// Since the last element of r is the unfactorized part,
	// it can only have a count of 1.
	assert(r.GetCount() == 1);
	zdiv(n1, *r.GetNum(), &f1, &temp);

	if (trace_level > 3)
	{
		cout << "  u = " << u;
		cout << ", r.GetNum() = " << *r.GetNum() << endl;
		cout << "  factors of " << n1 << ":";
		cout << r << " (";
		print_state(r.GetState());
		cout << ")" << endl;
	}

	zsq(f1, &temp);
	zsadd(temp, 1, &temp);
	zmulin(f1, &temp);
	z2mul(temp, &temp);	// temp = 2 * f1 * (f1^2 + 1)

	if (zscompare(f1, 2) <= 0 || zcompare(n, temp) >= 0)
	{
		if (trace_level > 1)
			cout << "  cannot decide" << endl;
		// XXX - here we could try harder to factorize the last element
		// of r. We know that it is composite or a probable prime.
		return NumProbPrime;
	}

	int t = 0;
	zuintoz(2, &base);

	while (1)
	{
		if (trace_level > 3)
			cout << "  Test base " << base << endl;

		zexpmod_m_ary(base, n1, n, &temp, 1);
		if (zscompare(temp, 1) > 0)
		{
			if (trace_level > 1)
				cout << "  composite because of Fermat test" << endl;
			return NumComposite;		// Fermat test
		}

		POSITION	pos = r.GetHeadPosition();
		bool		prime = true;

		while (pos != NULL)
		{
			IntElem* p = r.GetNext(pos);
			// The last iteration is unnecessary because at
			// this point n is already proven to be prime.
			if (pos == NULL)
				break;
			zdiv(n1, *p->GetNum(), &f2, &temp);
			zexpmod_m_ary(base, f2, n, &temp, 1);
			zsadd(temp, -1, &temp);
			if (zsign(temp) == 0)
				continue;
			zgcd(temp, n, factor);
			if (zscompare(*factor, 1) > 0)
			{
				prime = false;
				// base**(n-1)/q-1 could be divisible by n in which case
				// the gcd would become n.
				if (zcompare(*factor, n) < 0)
				{
					if (trace_level > 1)
						cout << "  Found a nontrivial factor: " << *factor << endl;
					return NumComposite;
				}
			}
		}

		if (prime)
		{
			// Final test.
			static verylong r1 = 0, s1 = 0;

			zdiv(*r.GetNum(), f1, &r1, &s1);

			if (zsmod(r1, 2) == 0)		// r1 even?
			{
				zsmul(r1, 4, &r1);
				zsq(s1, &base);		// base no longer needed
				zsub(s1, r1, &r1);
				if (zsqrt(r1, &temp, &base) == 1)
				{
					zsub(s1, temp, &s1);
					zmul(s1, f1, &temp);	// temp = (s1 - temp) * f1
					z2div(temp, &temp);
					zsadd(temp, 1, factor);
					if (trace_level > 1)
						cout << "  Found a nontrivial factor: " << *factor << endl;
					return NumComposite;
				}
			}

			if (trace_level > 1)
				cout << "theorem11 result: prime" << endl;

			return NumPrime;
		}

		if (t > 200)
		{
			if (trace_level > 1)
				cout << "  No base found, trying to find a factor..." << endl;
			return find_factor(n, u, factor);
		}

		zsadd(base, primeDiff[t++], &base);
	} // while

	assert(false);
	return 0; 		// Not reached

} // theorem11





ushort prove_prime(verylong n, ulong l, ulong u, verylong* f)
{
	IntList		r1;
	IntList		r2;
	IntList		r3;
	ushort		s = NumProbPrime;

	if (trace_level > 0)
		cout << "prove_prime: n = " << n << endl;

	if (l > 0)
	{
		s = num_status(n, l);
		if (s == NumPrime)
		{
			if (trace_level > 1)
				cout << "prove_prime result: prime" << endl;
			return s;	// Shortcut
		}
	}

	// Ignore lower bound l here because it is only valid for
	// n, not n-1 and n+1.
	factor_brute3(n, 0, u, r1, r2, r3);
	zcopy(*r2.GetNum(), f);
	if (zcompare(*f, n) != 0)
	{
		// Oops, found a factor! Maybe n is a pseudoprime.
		if (trace_level > 1)
			cout << "Oops, found a factor (" << *f
				<< ")! Maybe n is a pseudoprime." << endl;
		return NumComposite;
	}

	if (u > l)
	{
		s = num_status(n, u);
		if (s == NumPrime)
		{
			if (trace_level > 1)
				cout << "prove_prime result: prime" << endl;
			return s;
		}
	}

	assert(s == NumProbPrime);
	// Returns NumProbPrime if it can't decide; sets f to 1 if no factor
	// is found.
	s = theorem10(n, r1, r1.GetState(), u, f);

	// If the unfactorized part of n-1 is a probable prime, try to prove
	// it prime. If this succeeds the factorization of n-1 is complete and
	// we can apply theorem9 to prove n prime.
	if (s == NumProbPrime && r1.GetState() == NumProbPrime)
	{
		s = complete(r1, l, u);
		if (s == NumPrime)
			s = theorem9(n, r1, s, u, f);
	}

	if (s == NumProbPrime)
	{
		// theorem8 only works if n+1 is completely factored.
		if (r3.GetState() == NumProbPrime)
			complete(r3, l, u);
		if (r3.GetState() == NumPrime)
			s = theorem8(n, r3, s, u, f);
	}

	if (s == NumProbPrime)
	{
		if (trace_level > 1)
			cout << "prove_prime: No decision yet, trying brute-force recursion" << endl;

		verylong n1 = 0;		// Not static (recursion!)
		zcopy(*r1.GetNum(), &n1);

		// Can only fail under very weird circumstances which would
		// need additional research. Eg. if theorem9 or theorem10 can't
		// find a factor of a Carmichael number.
		assert(r1.GetCount() == 1);

		r1.Remove();
		IntList l1, l2;
		ulong u2;
		if (u > MaxUpperBound / 10)
			u2 = MaxUpperBound;
		else
			u2 = 10 * u;
		// We know n1 has no factor below u. We need an empty list
		// as input for factor so that add_cofactor works properly.
		s = factor(n1, u, u2, l2);
		l1.Merge(r1, l2);
		s = theorem10(n, l1, s, u2, f);

		zfree(&n1);
	}

	if (trace_level > 1)
	{
		cout << "prove_prime result: ";
		print_state(s);
		cout << endl;
	}

	return s;

} // prove_prime

