#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "base.h"

// Searches numbers whose factor sum is a certain distance greater than
// themselves.

const int MaxFactors = 64;

void root(verylong p, verylong start, ulong n, verylong* r)
{
	static verylong	old = 0;
	static verylong	temp = 0;
	static verylong	temp1 = 0;
	static verylong	temp2 = 0;

	zcopy(start, r);

	if (n == 2)
	{
		do
		{
			zcopy(*r, &old);
			zsq(old, &temp);
			zsub(temp, p, &temp);	// temp = old * old - p
			zsmul(old, n, &temp1);
			zdiv(temp, temp1, &temp, &temp1);
			// Formula: r = old - (old * old - p) / (old * n);
			zsub(old, temp, r);
			//cout << p << "**(1/2";
			//cout << ") = " << *r << endl;
		} while (zcompare(*r, old) != 0);
	}

	do
	{
		zcopy(*r, &old);
		zsexp(old, n - 1, &temp);
		zmul(old, temp, &temp2);
		zsub(temp2, p, &temp2);	// temp2 = old * temp - p
		zsmul(temp, n, &temp1);
		zdiv(temp2, temp1, &temp, &temp1);
		// Formula: r = old - (old * temp - p) / (temp * n);
		zsub(old, temp, r);
		//cout << p << "**(1/" << n;
		//cout << ") = " << *r << endl;
	} while (zcompare(*r, old) != 0);

}

void print(verylong p, verylong* d, ulong n)
{
	cout << p << " =";
	for (ulong j = 0; j < n; j++)
		cout << " " << d[j];
	cout << endl;
}

void NProd(ulong b, ulong n, int t, verylong* r)
{
	zone(r);

	if (trace_level > 5)
		cout << "NProd(" << b << ", " << n << ") = ";

	while (--n > 0)
	{
		zsmul(*r, b, r);	// Amicable numbers need not be sqarefree
	}

	if (trace_level > 5)
		cout << *r << endl;

}

// Calculate the factorsum of the number n whose prime factors are given
// in d[0]..d[n-1]. The sum includes n.
void fsum(ulong n, verylong* d, verylong* r)
{
	static verylong	temp = 0;
	ulong		i = 0;
	ulong		mark = 0;

	assert(n > 1);
	zone(r);

	while (i < n)
	{
		while (++i < n && zcompare(d[i - 1], d[i]) == 0)
			;
		if (i - mark == 1)
			zsadd(d[mark], 1, &temp);
		else
			powersum(d[mark], i - mark, &temp);
		zmulin(temp, r);
		mark = i;
	}
}

int Usage(char** argv, int errorCode)
{
	cout << "Usage: " << argv[0] << " start end distance" << endl;
	cout << "  start > 1" << endl;
	cout << "  end >= start" << endl;
	cout << "  distance > 0" << endl;
	cout << "Calculates all numbers between start and end whose factorsum"
		" is exactly distance greater than the number itself" << endl;
	return errorCode;
}

int main(int argc, char** argv)
{
	verylong	l = 0;
	verylong	u = 0;
	verylong	dist = 0;
	verylong	temp = 0;
	verylong	temp1 = 0;

	trace_level = 0;

	if (InitBase() == false)
		return 1;

	if (argc < 4)
		return Usage(argv, 1);

	if (zsread(argv[1], &l) == 0 || zsign(l) < 0)
		return Usage(argv, 1);

	if (zsread(argv[2], &u) == 0 || zsign(u) == 0 || zcompare(u, l) <= 0)
		return Usage(argv, 1);

	if (zsread(argv[3], &dist) == 0 || zsign(dist) <= 0)
		return Usage(argv, 1);

	//InitCheckMark(argv[0]);

	verylong	d[MaxFactors];
	verylong	lastD = 0;
	verylong	limit[MaxFactors];
	verylong	lastLimit = 0;
	verylong	p[MaxFactors];
	verylong	fs = 0;
	verylong	prevProd = 0;
	int		t[MaxFactors];
	int		lastT;
	verylong	lowerbound = 0;
	ulong		n = 3;
	ulong		i = 0;
	int		j;

	// One time initialization.
	for (j = 0; j < MaxFactors; j++)
	{
		d[j] = 0;
		limit[j] = 0;
		p[j] = 0;
	}

	do	// For each number of factors 3,4,...
	{
		//if (trace_level > 0)
			cout << "Searching for numbers with " << n << " factors" << endl;

		zuintoz(2, &d[0]);
		zuintoz(2, &p[0]);
		t[0] = 0;
		// Formula: limit[0] = u / NProd(d[0], n, t[0])
		NProd(ztouint(d[0]), n, t[0], &temp);
		zdiv(u, temp, &limit[0], &temp1);

		if (zscompare(limit[0], 2) < 0)
			break;

		do
		{
			while (++i < n - 1)
			{
				t[i] = t[i - 1];
				zcopy(d[i - 1], &d[i]);
				zmul(p[i - 1], d[i], &p[i]);
				NProd(ztouint(d[i]), n - i, t[i], &temp);
				zmul(p[i - 1], temp, &temp1);
				zdiv(u, temp1, &limit[i], &temp);
				assert(zscompare(limit[i], MaxUpperBound) <= 0);
			}

			zcopy(p[i - 1], &prevProd);
			zdiv(u, prevProd, &lastLimit, &temp);

			assert(zscompare(lastLimit, MaxUpperBound) <= 0);
			lastT = t[i - 1];
			zcopy(d[i - 1], &lastD);
			// Formula: lowerbound = (l + prevProd - 1) / prevProd;
			zadd(l, prevProd, &temp);
			zsdec(temp, 1);
			zdiv(temp, prevProd, &lowerbound, &temp1);
			if (trace_level > 5)
				cout << "lowerbound = " << lowerbound << endl;
			// Increase lastD until it equals or exceeds lowerbound.
			while (zcompare(lastD, lowerbound) < 0)
				zsinc(lastD, primeDiff[lastT++]);

			if (zcompare(lastD, d[i - 1]) == 0)
			{
				// Special case: the last two (or more) factors
				// are the same.
				zcopy(lastD, &d[i]);
				fsum(n, d, &fs);
				zdec(fs, dist);
				if (zeven(fs))
				{
					zrshift(fs, 1, &temp);
					if (zcompare(fs, p[i]) == 0)
					{
						cout << "Found one" << endl;
					}
				}
			}

			// Check that fs-dist is divisible by 2p[i-1]-fs
			fsum(i, d, &fs);
			zsub(fs, dist, &temp);
			zlshift(prevProd, 1, &temp1);
			zdec(temp1, fs);
			if (zsign(temp1) != 0)
			{
				zdiv(temp, temp1, &temp, &temp1);
				if (zsign(temp1) == 0
					&& zcompare(temp, lowerbound) > 0
					&& zcompare(temp, d[i - 1]) > 0
					&& zcompare(temp, lastLimit) <= 0
					&& prove_prime(temp, 0, MaxUpperBound, &temp1) == NumPrime)
				{
					zcopy(temp, &d[i]);
					zmul(prevProd, temp, &p[i]);
					print(p[i], d, n);
				}
			}

			do
			{
				i--;
				zsinc(d[i], primeDiff[t[i]++]);
				if (i > 0)
					zmul(d[i], p[i - 1], &p[i]);
				else
					zcopy(d[i], &p[i]);
			} while (i > 0 && zcompare(d[i], limit[i]) > 0);
		} while (zcompare(d[i], limit[i]) <= 0);

		n++;		// Increase number of factors
	} while (1);

	return 0;

}

